/**
 * Modulo: planificadorProcesos.c
 *
 * Descripcion: programa que simula un sencillo planificador de procesos
 * de un sistema operativo.
 *
 * Autor: Ivan Cantador, ivan.cantador@uam.es
 *
 * Version: 1.0
 * Fecha: 14/12/2016
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

// Definicion de valores constantes

#define NUM_PROCESOS 4

#define MILISECS_PER_SLOT 1000


// Definicion de tipos de datos

typedef enum {EN_ESPERA, EN_EJECUCION, COMPLETADO, BLOQUEADO} Estado;

typedef enum {CIRCULAR, CON_PRIORIDADES} EstrategiaPlanificacion;

typedef struct {
	int id;
	Estado estado;
	int prioridad;
	int tiempoAsignado;
} Proceso;

typedef struct {
	Proceso procesos[NUM_PROCESOS];
	int actual;
	int duracionSlot;
} TablaPlanifacionProcesos;


// Definicion de funciones

/**
 * Funcion: aleatorio
 *
 * Descripcion: genera un numero aleatorio uniformemente distribuido en un
 * intervalo [a, b]
 *
 * Argumentos de entrada:
 * - a: extremo izquierdo del intervalo
 * - b: extremo derecho del intervalo
 *
 * Retorno:
 * - Numero aleatorio generado
 */
int aleatorio(int a, int b) {
	return a + rand() * (b - a) / (RAND_MAX+1);
}

/**
 * Funcion: nombreEstado
 *
 * Descripcion: devuelve el nombre de un estado del planificador de procesos
 *
 * Argumentos de entrada:
 * - codigo: codigo del estado cuyo nombre hay que devolver
 * - nombre: cadena de caracteres donde guardar el nombre del estado
 *
 * Retorno:
 * - (nada)
 */
void nombreEstado(Estado codigo, char *nombre) {
	switch ( codigo ) {
		case EN_ESPERA:
			strcpy(nombre, "en espera");
			break;
		case EN_EJECUCION:
			strcpy(nombre, "en ejecucion");
			break;
		case COMPLETADO:
			strcpy(nombre, "completada");
			break;
		case BLOQUEADO:
			strcpy(nombre, "bloqueada");
			break;
		default:
			strcpy(nombre, "");
	}
}

/**
 * Funcion: asignacionCicular
 *
 * Descripcion: asigna un slot de tiempo de ejecucion a un proceso del
 * planificador siguiendo una asignacion de tipo circular.
 *
 * Argumentos de entrada:
 * - tabla: planificador de procesos
 *
 * Retorno:
 * - 0 si la asignacion se realiza con exito; -1 en caso contrario
 */
int asignacionCicular(TablaPlanifacionProcesos *tabla) {
	int i;

	if( tabla == NULL || tabla->actual < 0 || tabla->actual >= NUM_PROCESOS ) {
		return -1;
	}

	// El estado del proceso actual pasa a EN_ESPERA
	if( tabla->procesos[tabla->actual].estado == EN_EJECUCION ) {
		tabla->procesos[tabla->actual].estado = EN_ESPERA;
	}

	// Se busca la primera proceso EN_ESPERA de forma circular y se pone EN_EJECUCION
	for( i=0; i<NUM_PROCESOS; i++ ) {
		tabla->actual = (tabla->actual + 1) % NUM_PROCESOS;  // incremento circular
		if( tabla->procesos[tabla->actual].estado == EN_ESPERA ) {
			tabla->procesos[tabla->actual].estado = EN_EJECUCION;
			tabla->procesos[tabla->actual].tiempoAsignado += tabla->duracionSlot;
			return tabla->procesos[tabla->actual].id;
		}
	}

	return 0;
}

/**
 * Funcion: asignacionPorPrioridad
 *
 * Descripcion: asigna un slot de tiempo de ejecucion a un proceso del
 * planificador siguiendo una asignacion atendiendo a las prioridades
 * de los procesos.
 *
 * Argumentos de entrada:
 * - tabla: planificador de procesos
 *
 * Retorno:
 * - 0 si la asignacion se realiza con exito; -1 en caso contrario
 */
int asignacionPorPrioridad(TablaPlanifacionProcesos *tabla) {
	int i, t, maxPrioridad;

	if( tabla == NULL || tabla->actual < 0 || tabla->actual >= NUM_PROCESOS ) {
		return -1;
	}

	// Se busca la prioridad mas alta entre los procesos EN_ESPERA
	maxPrioridad = -1;
	t = aleatorio(0, NUM_PROCESOS-1);
	for( i=0; i<NUM_PROCESOS; i++ ) {
		if( tabla->procesos[(t+i)%NUM_PROCESOS].estado == EN_ESPERA ) {
			if( tabla->procesos[(t+i)%NUM_PROCESOS].prioridad > maxPrioridad ) {
				maxPrioridad = tabla->procesos[(t+i)%NUM_PROCESOS].prioridad;
			}
		}
	}

	// El estado del proceso actual pasa a EN_ESPERA
	if( tabla->procesos[tabla->actual].estado == EN_EJECUCION ) {
		tabla->procesos[tabla->actual].estado = EN_ESPERA;
	}

	// Se busca la primera proceso EN_ESPERA con prioridad mxima y se pone EN_EJECUCION
	for( i=0; i<NUM_PROCESOS; i++ ) {
		tabla->actual = (tabla->actual + 1) % NUM_PROCESOS;  // incremento circular
		if( tabla->procesos[tabla->actual].estado == EN_ESPERA && tabla->procesos[tabla->actual].prioridad == maxPrioridad ) {
			tabla->procesos[tabla->actual].estado = EN_EJECUCION;
			tabla->procesos[tabla->actual].tiempoAsignado += tabla->duracionSlot;
			tabla->procesos[tabla->actual].prioridad = (tabla->procesos[tabla->actual].prioridad - 1) % NUM_PROCESOS;
			return tabla->procesos[tabla->actual].id;
		}
	}

	return -1;
}

/**
 * Funcion: planificador
 *
 * Descripcion: funcion principal que de forma iterativa asigna slots de
 * tiempo de ejecucion a procesos registradas, atendiendo a una estrategia
 * (e.g., circular, por prioridades) dada.
 *
 * Argumentos de entrada:
 * - tabla: planificador de procesos
 * - estrategia: estrategia de asignacion de slots de tiempo de ejecucion
 * - numeroSlots: numero de slots a asignar entre los procesos
 *   (a quitar en version completa donde la tabla de procesos es compartida
 *   entre procesos y donde planificador se ejecuta ininterrumpidamente)
 *
 * Retorno:
 * - (nada)
 */
void planificador(TablaPlanifacionProcesos *tabla, EstrategiaPlanificacion estrategia, int numeroSlots) {
	char nombre[32];
	clock_t tIni, tFin;
	int i, t;

	if( tabla == NULL || (estrategia != CIRCULAR && estrategia != CON_PRIORIDADES) ) {
		return;
	}

	for ( i=0; i<numeroSlots; i++ ) {
		printf("%d: Proceso actual: %d.\n", i+1, tabla->actual);

		for( i=0; i<NUM_PROCESOS; i++ ) {
			nombreEstado(tabla->procesos[i].estado, nombre);
			printf("\tProceso %d (p=%d): %s.\n", tabla->procesos[i].id, tabla->procesos[i].prioridad, nombre);
		}

		tIni = clock();
		do {
			tFin = clock();
			t = (int) (((double) (tFin - tIni) / CLOCKS_PER_SEC) * 1000);
		} while( t < MILISECS_PER_SLOT );

		switch( estrategia ) {
			case CIRCULAR:
				asignacionCicular(tabla);
				break;
			case CON_PRIORIDADES:
				asignacionPorPrioridad(tabla);
				break;
		}
	} // while
}

/**
 * Funcion: main
 *
 * Descripcion: programa que simula la asignacion de slots de tiempo de ejecucion
 * a un conjunto de procesos, siguiendo una estrategia (circular, con prioridades)
 * concrteta durante un tiempo (numero de slots) dado.
 * En un entorno real la ejecucion del programa es ininterrumpida, los procesos se
 * reemplazan por otras cuando han acabado de ejecucion, y sus estados pueden
 * alterarse de manera externa, e.g., pasando al estado de bloqueo.
 *
 * Argumentos de entrada:
 * - (ninguno)
 *
 * Retorno:
 * - (ninguno)
 */
int main() {
	TablaPlanifacionProcesos tablaProcesos;
	int i;

	// Inicializamos el generador de numeros aleatorios
	srand((unsigned) time(NULL));

	for( i=0; i<NUM_PROCESOS; i++ ) {
		tablaProcesos.procesos[i].id = i;
		tablaProcesos.procesos[i].estado = EN_ESPERA;
		tablaProcesos.procesos[i].prioridad = aleatorio(1, 3);
		tablaProcesos.procesos[i].tiempoAsignado = 0;
	}
	tablaProcesos.actual = 0;
	tablaProcesos.duracionSlot = MILISECS_PER_SLOT;

	planificador(&tablaProcesos, CON_PRIORIDADES, 10);

	return 1;
}
